const std = @import("std");
const mem = std.mem;

const xml = @import("xml.zig");

const gpa = &allocator_instance.allocator;
var allocator_instance = std.heap.GeneralPurposeAllocator(.{}){};

pub fn scan(out_dir: std.fs.Dir, wayland_xml: []const u8, protocols: []const []const u8) !void {
    const wayland_file = try out_dir.createFile("wayland.zig", .{});
    try wayland_file.writeAll(@embedFile("wayland.zig"));
    defer wayland_file.close();

    var scanner = Scanner{};

    try scanner.scanProtocol(out_dir, wayland_xml);
    for (protocols) |xml_filename|
        try scanner.scanProtocol(out_dir, xml_filename);

    {
        const client_core_file = try out_dir.createFile("wayland_client_core.zig", .{});
        defer client_core_file.close();
        try client_core_file.writeAll(@embedFile("wayland_client_core.zig"));

        const client_file = try out_dir.createFile("client.zig", .{});
        defer client_file.close();
        const writer = client_file.writer();

        var iter = scanner.client.iterator();
        while (iter.next()) |entry| {
            try writer.print("pub const {} = struct {{", .{entry.key});
            if (mem.eql(u8, entry.key, "wl"))
                try writer.writeAll("pub usingnamespace @import(\"wayland_client_core.zig\");\n");
            for (entry.value.items) |generated_file|
                try writer.print("pub usingnamespace @import(\"{}\");", .{generated_file});
            try writer.writeAll("};\n");
        }
    }

    {
        const server_core_file = try out_dir.createFile("wayland_server_core.zig", .{});
        defer server_core_file.close();
        try server_core_file.writeAll(@embedFile("wayland_server_core.zig"));

        const server_file = try out_dir.createFile("server.zig", .{});
        defer server_file.close();
        const writer = server_file.writer();

        var iter = scanner.server.iterator();
        while (iter.next()) |entry| {
            try writer.print("pub const {} = struct {{", .{entry.key});
            if (mem.eql(u8, entry.key, "wl"))
                try writer.writeAll("pub usingnamespace @import(\"wayland_server_core.zig\");\n");
            for (entry.value.items) |generated_file|
                try writer.print("pub usingnamespace @import(\"{}\");", .{generated_file});
            try writer.writeAll("};\n");
        }
    }

    {
        const common_file = try out_dir.createFile("common.zig", .{});
        defer common_file.close();
        const writer = common_file.writer();
        try writer.writeAll(@embedFile("common_core.zig"));

        var iter = scanner.common.iterator();
        while (iter.next()) |entry| {
            try writer.print("pub const {} = struct {{", .{entry.key});
            for (entry.value.items) |generated_file|
                try writer.print("pub usingnamespace @import(\"{}\");", .{generated_file});
            try writer.writeAll("};\n");
        }
    }
}

const Side = enum {
    client,
    server,
};

const Scanner = struct {
    /// Map from namespace to list of generated files
    const Map = std.hash_map.StringHashMap(std.ArrayListUnmanaged([]const u8));
    client: Map = Map.init(gpa),
    server: Map = Map.init(gpa),
    common: Map = Map.init(gpa),

    fn scanProtocol(scanner: *Scanner, out_dir: std.fs.Dir, xml_filename: []const u8) !void {
        const xml_file = try std.fs.openFileAbsolute(xml_filename, .{});
        defer xml_file.close();

        var arena = std.heap.ArenaAllocator.init(gpa);
        defer arena.deinit();

        const xml_bytes = try xml_file.readToEndAlloc(&arena.allocator, 512 * 4096);
        const protocol = try Protocol.parseXML(&arena.allocator, xml_bytes);

        const xml_basename = std.fs.path.basename(xml_filename);
        const protocol_name = try gpa.dupe(u8, xml_basename[0 .. xml_basename.len - 4]);
        for (protocol_name) |*ch| {
            if (ch.* == '-') ch.* = '_';
        }
        const protocol_namespace = try gpa.dupe(u8, protocol.namespace);

        {
            const client_filename = try mem.concat(gpa, u8, &[_][]const u8{ protocol_name, "_client.zig" });
            const client_file = try out_dir.createFile(client_filename, .{});
            defer client_file.close();
            try protocol.emitClient(client_file.writer());
            try (try scanner.client.getOrPutValue(protocol_namespace, .{})).value.append(gpa, client_filename);
        }

        {
            const server_filename = try mem.concat(gpa, u8, &[_][]const u8{ protocol_name, "_server.zig" });
            const server_file = try out_dir.createFile(server_filename, .{});
            defer server_file.close();
            try protocol.emitServer(server_file.writer());
            try (try scanner.server.getOrPutValue(protocol_namespace, .{})).value.append(gpa, server_filename);
        }

        {
            const common_filename = try mem.concat(gpa, u8, &[_][]const u8{ protocol_name, "_common.zig" });
            const common_file = try out_dir.createFile(common_filename, .{});
            defer common_file.close();
            try protocol.emitCommon(common_file.writer());
            try (try scanner.common.getOrPutValue(protocol_namespace, .{})).value.append(gpa, common_filename);
        }
    }
};

const Protocol = struct {
    name: []const u8,
    namespace: []const u8,
    interfaces: std.ArrayList(Interface),

    fn parseXML(allocator: *mem.Allocator, xml_bytes: []const u8) !Protocol {
        var parser = xml.Parser.init(xml_bytes);
        while (parser.next()) |ev| switch (ev) {
            .open_tag => |tag| if (mem.eql(u8, tag, "protocol")) return parse(allocator, &parser),
            else => {},
        };
        return error.UnexpectedEndOfFile;
    }

    fn parse(allocator: *mem.Allocator, parser: *xml.Parser) !Protocol {
        var name: ?[]const u8 = null;
        var interfaces = std.ArrayList(Interface).init(allocator);
        while (parser.next()) |ev| switch (ev) {
            .open_tag => |tag| {
                // TODO: parse copyright and description
                if (mem.eql(u8, tag, "interface"))
                    try interfaces.append(try Interface.parse(allocator, parser));
            },
            .attribute => |attr| if (mem.eql(u8, attr.name, "name")) {
                if (name != null) return error.DuplicateName;
                name = try attr.dupeValue(allocator);
            },
            .close_tag => |tag| if (mem.eql(u8, tag, "protocol")) {
                return Protocol{
                    .name = name orelse return error.MissingName,
                    .namespace = if (mem.eql(u8, name.?, "wayland")) "wl" else prefix(name.?),
                    .interfaces = interfaces,
                };
            },
            else => {},
        };
        return error.UnexpectedEndOfFile;
    }

    fn emitClient(protocol: Protocol, writer: anytype) !void {
        try writer.writeAll(
            \\const os = @import("std").os;
            \\const client = @import("wayland.zig").client;
            \\const common = @import("common.zig");
        );
        for (protocol.interfaces.items) |interface|
            try interface.emit(.client, protocol.namespace, writer);
    }

    fn emitServer(protocol: Protocol, writer: anytype) !void {
        try writer.writeAll(
            \\const os = @import("std").os;
            \\const server = @import("wayland.zig").server;
            \\const common = @import("common.zig");
        );
        for (protocol.interfaces.items) |interface|
            try interface.emit(.server, protocol.namespace, writer);
    }

    fn emitCommon(protocol: Protocol, writer: anytype) !void {
        try writer.writeAll(
            \\const common = @import("common.zig");
        );
        for (protocol.interfaces.items) |interface|
            try interface.emitCommon(writer);
    }
};

const Interface = struct {
    name: []const u8,
    version: u32,
    requests: std.ArrayList(Message),
    events: std.ArrayList(Message),
    enums: std.ArrayList(Enum),

    fn parse(allocator: *mem.Allocator, parser: *xml.Parser) !Interface {
        var name: ?[]const u8 = null;
        var version: ?u32 = null;
        var requests = std.ArrayList(Message).init(allocator);
        var events = std.ArrayList(Message).init(allocator);
        var enums = std.ArrayList(Enum).init(allocator);

        while (parser.next()) |ev| switch (ev) {
            .open_tag => |tag| {
                // TODO: parse description
                if (mem.eql(u8, tag, "request"))
                    try requests.append(try Message.parse(allocator, parser))
                else if (mem.eql(u8, tag, "event"))
                    try events.append(try Message.parse(allocator, parser))
                else if (mem.eql(u8, tag, "enum"))
                    try enums.append(try Enum.parse(allocator, parser));
            },
            .attribute => |attr| {
                if (mem.eql(u8, attr.name, "name")) {
                    if (name != null) return error.DuplicateName;
                    name = try attr.dupeValue(allocator);
                } else if (mem.eql(u8, attr.name, "version")) {
                    if (version != null) return error.DuplicateVersion;
                    version = try std.fmt.parseInt(u32, try attr.dupeValue(allocator), 10);
                }
            },
            .close_tag => |tag| if (mem.eql(u8, tag, "interface")) {
                return Interface{
                    .name = name orelse return error.MissingName,
                    .version = version orelse return error.MissingVersion,
                    .requests = requests,
                    .events = events,
                    .enums = enums,
                };
            },
            else => {},
        };
        return error.UnexpectedEndOfFile;
    }

    fn emit(interface: Interface, side: Side, namespace: []const u8, writer: anytype) !void {
        var buf: [1024]u8 = undefined;
        var fbs = std.io.fixedBufferStream(&buf);
        try printIdentifier(fbs.writer(), case(.title, trimPrefix(interface.name)));
        const title_case = fbs.getWritten();
        try printIdentifier(fbs.writer(), trimPrefix(interface.name));
        const snake_case = fbs.getWritten()[title_case.len..];

        try writer.print(
            \\pub const {} = opaque {{
            \\ pub const interface = common.{}.{}.interface;
        , .{ title_case, namespace, snake_case });

        for (interface.enums.items) |e| {
            try writer.writeAll("pub const ");
            try printIdentifier(writer, case(.title, e.name));
            try writer.print(" = common.{}.{}.", .{ namespace, snake_case });
            try printIdentifier(writer, case(.title, e.name));
            try writer.writeAll(";\n");
        }

        if (side == .client) {
            if (interface.events.items.len > 0) {
                try writer.writeAll("pub const Event = union(enum) {");
                for (interface.events.items) |event| try event.emitField(.client, writer);
                try writer.writeAll("};\n");
                try writer.print(
                    \\pub fn setListener(
                    \\    {}: *{},
                    \\    comptime T: type,
                    \\    listener: fn ({}: *{}, event: Event, data: T) void,
                    \\    data: T,
                    \\) !void {{
                    \\    const proxy = @ptrCast(*client.wl.Proxy, {});
                    \\    try proxy.addDispatcher(common.Dispatcher({}, T).dispatcher, listener, data);
                    \\}}
                , .{ snake_case, title_case, snake_case, title_case, snake_case, title_case });
            }

            for (interface.requests.items) |request, opcode|
                try request.emitFn(side, writer, interface, opcode);

            if (mem.eql(u8, interface.name, "wl_display"))
                try writer.writeAll(@embedFile("client_display_functions.zig"));
        } else {
            try writer.print(
                \\pub fn create(client: *server.wl.Client, version: u32, id: u32) !*{} {{
                \\    return @ptrCast(*{}, try server.wl.Resource.create(client, {}, version, id));
                \\}}
            , .{ title_case, title_case, title_case });

            for ([_][]const u8{ "destroy", "getId", "fromLink", "getLink", "getClient", "getVersion" }) |func|
                try writer.print(
                    \\pub fn {}({}: *{}) void {{
                    \\    @ptrCast(*server.wl.Resource, {}).{}();
                    \\}}
                , .{ func, snake_case, title_case, snake_case, func });
            if (interface.requests.items.len > 0) {
                try writer.writeAll("pub const Request = union(enum) {");
                for (interface.requests.items) |request| try request.emitField(.server, writer);
                try writer.writeAll("};\n");
                @setEvalBranchQuota(2100);
                try writer.print(
                    \\pub fn setHandler(
                    \\    {}: *{},
                    \\    comptime T: type,
                    \\    handler: fn ({}: *{}, request: Request, data: T) void,
                    \\    destroy: fn ({}: *{}, data: T) void,
                    \\    data: T,
                    \\) void {{
                    \\    const resource = @ptrCast(*server.wl.Resource, {});
                    \\    resource.setDispatcher(
                    \\        common.Dispatcher({}, T).dispatcher,
                    \\        handler,
                    \\        data,
                    \\        struct {{
                    \\            fn wrapper(resource: *server.wl.Resource) callconv(.C) void {{
                    \\                @call(.{{ .modifier = .always_inline }}, destroy, .{{
                    \\                    @ptrCast(*{}, resource),
                    \\                    @ptrCast(T, resource.getUserData()),
                    \\                }});
                    \\            }}
                    \\        }}.wrapper,
                    \\    );
                    \\}}
                , .{ snake_case, title_case, snake_case, title_case, snake_case, title_case, snake_case, title_case, title_case });
            } else {
                try writer.print(
                    \\pub fn setHandler(
                    \\    {}: *{},
                    \\    comptime T: type,
                    \\    destroy: fn ({}: *{}, data: T) void,
                    \\    data: T,
                    \\) void {{
                    \\    const resource = @ptrCast(*server.wl.Resource, {});
                    \\    resource.setDispatcher(
                    \\        null,
                    \\        null,
                    \\        data,
                    \\        struct {{
                    \\            fn wrapper(resource: *server.wl.Resource) callconv(.C) void {{
                    \\                @call(.{{ .modifier = .always_inline }}, destroy, .{{
                    \\                    @ptrCast(*{}, resource),
                    \\                    @ptrCast(T, resource.getUserData()),
                    \\                }});
                    \\            }}
                    \\        }}.wrapper,
                    \\    );
                    \\}}
                , .{ snake_case, title_case, snake_case, title_case, snake_case, title_case });
            }

            for (interface.events.items) |event, opcode|
                try event.emitFn(side, writer, interface, opcode);
        }

        try writer.writeAll("};\n");
    }

    fn emitCommon(interface: Interface, writer: anytype) !void {
        try writer.writeAll("pub const ");
        try printIdentifier(writer, trimPrefix(interface.name));

        // TODO: stop linking libwayland generated interface structs when
        // https://github.com/ziglang/zig/issues/131 is implemented
        //
        //try writer.print(
        //    \\ = struct {{
        //    \\ pub const interface = common.Interface{{
        //    \\  .name = "{}",
        //    \\  .version = {},
        //    \\  .method_count = {},
        //, .{
        //    interface.name,
        //    interface.version,
        //    interface.requests.items.len,
        //});
        //if (interface.requests.items.len > 0) {
        //    try writer.writeAll(".methods = &[_]common.Message{");
        //    for (interface.requests.items) |request| try request.emitMessage(writer);
        //    try writer.writeAll("},");
        //} else {
        //    try writer.writeAll(".methods = null,");
        //}
        //try writer.print(".event_count = {},", .{interface.events.items.len});
        //if (interface.events.items.len > 0) {
        //    try writer.writeAll(".events = &[_]common.Message{");
        //    for (interface.events.items) |event| try event.emitMessage(writer);
        //    try writer.writeAll("},");
        //} else {
        //    try writer.writeAll(".events = null,");
        //}
        //try writer.writeAll("};");

        try writer.print(
            \\ = struct {{
            \\ extern const {}_interface: common.Interface;
            \\ pub inline fn interface() *const common.Interface {{
            \\  return &{}_interface;
            \\ }}
        , .{ interface.name, interface.name });

        for (interface.enums.items) |e| try e.emit(writer);
        try writer.writeAll("};");
    }
};

const Message = struct {
    name: []const u8,
    since: u32,
    args: std.ArrayList(Arg) = std.ArrayList(Arg),
    kind: union(enum) {
        normal: void,
        constructor: ?[]const u8,
        destructor: void,
    },

    fn parse(allocator: *mem.Allocator, parser: *xml.Parser) !Message {
        var name: ?[]const u8 = null;
        var since: ?u32 = null;
        var args = std.ArrayList(Arg).init(allocator);
        var destructor = false;

        while (parser.next()) |ev| switch (ev) {
            .open_tag => |tag| {
                // TODO: parse description
                if (mem.eql(u8, tag, "arg"))
                    try args.append(try Arg.parse(allocator, parser));
            },
            .attribute => |attr| {
                if (mem.eql(u8, attr.name, "name")) {
                    if (name != null) return error.DuplicateName;
                    name = try attr.dupeValue(allocator);
                } else if (mem.eql(u8, attr.name, "since")) {
                    if (since != null) return error.DuplicateSince;
                    since = try std.fmt.parseInt(u32, try attr.dupeValue(allocator), 10);
                } else if (mem.eql(u8, attr.name, "type")) {
                    if (attr.valueEql("destructor")) {
                        destructor = true;
                    } else {
                        return error.InvalidType;
                    }
                }
            },
            .close_tag => |tag| if (mem.eql(u8, tag, "request") or mem.eql(u8, tag, "event")) {
                return Message{
                    .name = name orelse return error.MissingName,
                    .since = since orelse 1,
                    .args = args,
                    .kind = blk: {
                        if (destructor) break :blk .destructor;
                        for (args.items) |arg|
                            if (arg.kind == .new_id) break :blk .{ .constructor = arg.kind.new_id };
                        break :blk .normal;
                    },
                };
            },
            else => {},
        };
        return error.UnexpectedEndOfFile;
    }

    fn emitMessage(message: Message, writer: anytype) !void {
        try writer.print(".{{ .name = \"{}\", .signature = \"", .{message.name});
        for (message.args.items) |arg| try arg.emitSignature(writer);
        try writer.writeAll("\", .types = ");
        if (message.args.items.len > 0) {
            try writer.writeAll("&[_]?*const common.Interface{");
            for (message.args.items) |arg| {
                switch (arg.kind) {
                    .object, .new_id => |interface| if (interface) |i|
                        try writer.print("&common.{}.{}.interface,", .{ prefix(i), trimPrefix(i) })
                    else
                        try writer.writeAll("null,"),
                    else => try writer.writeAll("null,"),
                }
            }
            try writer.writeAll("},");
        } else {
            try writer.writeAll("null,");
        }
        try writer.writeAll("},");
    }

    fn emitField(message: Message, side: Side, writer: anytype) !void {
        try printIdentifier(writer, message.name);
        try writer.writeAll(": struct {");
        for (message.args.items) |arg| {
            if (side == .server and arg.kind == .new_id and arg.kind.new_id == null) {
                try writer.writeAll("interface: [*:0]const u8, version: u32,");
                try printIdentifier(writer, arg.name);
                try writer.writeAll(": u32");
            } else if (side == .client and arg.kind == .new_id) {
                try printIdentifier(writer, arg.name);
                try writer.writeAll(": *");
                try printAbsolute(.client, writer, arg.kind.new_id.?);
                std.debug.assert(!arg.allow_null);
            } else {
                try printIdentifier(writer, arg.name);
                try writer.writeByte(':');
                // See notes on NULL in doc comment for wl_message in wayland-util.h
                if (side == .client and arg.kind == .object and !arg.allow_null)
                    try writer.writeByte('?');
                try arg.emitType(side, writer);
            }
            try writer.writeByte(',');
        }
        try writer.writeAll("},\n");
    }

    fn emitFn(message: Message, side: Side, writer: anytype, interface: Interface, opcode: usize) !void {
        try writer.writeAll("pub fn ");
        if (side == .server) {
            try writer.writeAll("send");
            try printIdentifier(writer, case(.title, message.name));
        } else {
            try printIdentifier(writer, case(.camel, message.name));
        }
        try writer.writeByte('(');
        try printIdentifier(writer, trimPrefix(interface.name));
        try writer.writeAll(": *");
        try printIdentifier(writer, case(.title, trimPrefix(interface.name)));
        for (message.args.items) |arg| {
            if (side == .server and arg.kind == .new_id) {
                try writer.writeByte(',');
                try printIdentifier(writer, arg.name);
                try writer.writeByte(':');
                if (arg.allow_null) try writer.writeByte('?');
                try writer.writeByte('*');
                if (arg.kind.new_id) |iface|
                    try printIdentifier(writer, case(.title, trimPrefix(iface)))
                else
                    try writer.writeAll("server.wl.Resource");
            } else if (side == .client and arg.kind == .new_id) {
                if (arg.kind.new_id == null) try writer.writeAll(", comptime T: type, version: u32");
            } else {
                try writer.writeByte(',');
                try printIdentifier(writer, arg.name);
                try writer.writeByte(':');
                try arg.emitType(side, writer);
            }
        }
        if (side == .server or message.kind != .constructor) {
            try writer.writeAll(") void {");
        } else if (message.kind.constructor) |new_iface| {
            try writer.writeAll(") !*");
            try printIdentifier(writer, case(.title, trimPrefix(new_iface)));
            try writer.writeAll("{");
        } else {
            try writer.writeAll(") !*T {");
        }
        if (side == .server)
            try writer.writeAll("const resource = @ptrCast(*server.wl.Resource,")
        else
            try writer.writeAll("const proxy = @ptrCast(*client.wl.Proxy,");
        try printIdentifier(writer, trimPrefix(interface.name));
        try writer.writeAll(");");
        if (message.args.items.len > 0) {
            try writer.writeAll("var args = [_]common.Argument{");
            for (message.args.items) |arg| {
                switch (arg.kind) {
                    .int, .uint, .fixed, .string, .array, .fd => {
                        try writer.writeAll(".{ .");
                        try arg.emitSignature(writer);
                        try writer.writeAll(" = ");
                        if (arg.enum_name != null) {
                            try writer.writeAll("switch (@typeInfo(");
                            try arg.emitType(side, writer);

                            // TODO We know the type of the enum at scanning time, but it's
                            //      currently a bit difficult to access it.
                            const c_type = if (arg.kind == .uint) "u32" else "i32";
                            try writer.print(
                                \\ )) {{
                                \\    .Enum => @intCast({}, @enumToInt({})),
                                \\    .Struct => @bitCast(u32, {}),
                                \\    else => unreachable,
                                \\ }}
                            , .{ c_type, arg.name, arg.name });
                        } else {
                            try printIdentifier(writer, arg.name);
                        }
                        try writer.writeAll("},");
                    },
                    .object, .new_id => |new_iface| {
                        if (arg.kind == .object or side == .server) {
                            if (arg.allow_null) {
                                try writer.writeAll(".{ .o = if (");
                                try printIdentifier(writer, arg.name);
                                try writer.writeAll(") |o| @ptrCast(*common.Object, o) else null },");
                            } else {
                                try writer.writeAll(".{ .o = @ptrCast(*common.Object, ");
                                try printIdentifier(writer, arg.name);
                                try writer.writeAll(") },");
                            }
                        } else {
                            if (new_iface == null) {
                                try writer.writeAll(
                                    \\.{ .s = T.interface().name },
                                    \\.{ .u = version },
                                );
                            }
                            try writer.writeAll(".{ .o = null },");
                        }
                    },
                }
            }
            try writer.writeAll("};\n");
        }
        const args = if (message.args.items.len > 0) "&args" else "null";
        if (side == .server) {
            try writer.print("resource.postEvent({}, {});", .{ opcode, args });
        } else switch (message.kind) {
            .normal, .destructor => {
                try writer.print("proxy.marshal({}, {});", .{ opcode, args });
                if (message.kind == .destructor) try writer.writeAll("proxy.destroy();");
            },
            .constructor => |new_iface| {
                if (new_iface) |i| {
                    try writer.writeAll("return @ptrCast(*");
                    try printIdentifier(writer, case(.title, trimPrefix(i)));
                    try writer.print(", try proxy.marshalConstructor({}, &args, ", .{opcode});
                    try printIdentifier(writer, case(.title, trimPrefix(i)));
                    try writer.writeAll(".interface()));");
                } else {
                    try writer.print("return @ptrCast(*T, try proxy.marshalConstructorVersioned({}, &args, T.interface(), version));", .{opcode});
                }
            },
        }
        try writer.writeAll("}\n");
    }
};

const Arg = struct {
    const Type = union(enum) {
        int,
        uint,
        fixed,
        string,
        new_id: ?[]const u8,
        object: ?[]const u8,
        array,
        fd,
    };
    name: []const u8,
    kind: Type,
    allow_null: bool,
    enum_name: ?[]const u8,

    fn parse(allocator: *mem.Allocator, parser: *xml.Parser) !Arg {
        var name: ?[]const u8 = null;
        var kind: ?@TagType(Type) = null;
        var interface: ?[]const u8 = null;
        var allow_null: ?bool = null;
        var enum_name: ?[]const u8 = null;

        while (parser.next()) |ev| switch (ev) {
            .attribute => |attr| {
                if (mem.eql(u8, attr.name, "name")) {
                    if (name != null) return error.DuplicateName;
                    name = try attr.dupeValue(allocator);
                } else if (mem.eql(u8, attr.name, "type")) {
                    if (kind != null) return error.DuplicateType;
                    kind = std.meta.stringToEnum(@TagType(Type), try attr.dupeValue(allocator)) orelse
                        return error.InvalidType;
                } else if (mem.eql(u8, attr.name, "interface")) {
                    if (interface != null) return error.DuplicateInterface;
                    interface = try attr.dupeValue(allocator);
                } else if (mem.eql(u8, attr.name, "allow-null")) {
                    if (allow_null != null) return error.DuplicateAllowNull;
                    if (!attr.valueEql("true") and !attr.valueEql("false")) return error.InvalidBoolValue;
                    allow_null = attr.valueEql("true");
                } else if (mem.eql(u8, attr.name, "enum")) {
                    if (enum_name != null) return error.DuplicateEnum;
                    enum_name = try attr.dupeValue(allocator);
                }
            },
            .close_tag => |tag| if (mem.eql(u8, tag, "arg")) {
                return Arg{
                    .name = name orelse return error.MissingName,
                    .kind = switch (kind orelse return error.MissingType) {
                        .object => .{ .object = interface },
                        .new_id => .{ .new_id = interface },
                        .int => .int,
                        .uint => .uint,
                        .fixed => .fixed,
                        .string => .string,
                        .array => .array,
                        .fd => .fd,
                    },
                    .allow_null = allow_null orelse false,
                    .enum_name = enum_name,
                };
            },
            else => {},
        };
        return error.UnexpectedEndOfFile;
    }

    fn emitSignature(arg: Arg, writer: anytype) !void {
        switch (arg.kind) {
            .int => try writer.writeByte('i'),
            .uint => try writer.writeByte('u'),
            .fixed => try writer.writeByte('f'),
            .string => try writer.writeByte('s'),
            .new_id => |interface| if (interface == null)
                try writer.writeAll("sun")
            else
                try writer.writeByte('n'),
            .object => try writer.writeByte('o'),
            .array => try writer.writeByte('a'),
            .fd => try writer.writeByte('h'),
        }
    }

    fn emitType(arg: Arg, side: Side, writer: anytype) !void {
        switch (arg.kind) {
            .int, .uint => {
                if (arg.enum_name) |name| {
                    if (mem.indexOfScalar(u8, name, '.')) |dot_index| {
                        // Turn a reference like wl_shm.format into common.wl.shm.Format
                        try writer.writeAll("common.");
                        const us_index = mem.indexOfScalar(u8, name, '_') orelse 0;
                        try writer.writeAll(name[0..us_index]);
                        try writer.writeAll(".");
                        try writer.writeAll(name[us_index + 1 .. dot_index + 1]);
                        try writer.writeAll(case(.title, name[dot_index + 1 ..]));
                    } else {
                        try writer.writeAll(case(.title, name));
                    }
                } else if (arg.kind == .int) {
                    try writer.writeAll("i32");
                } else {
                    try writer.writeAll("u32");
                }
            },
            .new_id => try writer.writeAll("u32"),
            .fixed => try writer.writeAll("common.Fixed"),
            .string => {
                if (arg.allow_null) try writer.writeByte('?');
                try writer.writeAll("[*:0]const u8");
            },
            .object => |interface| if (interface) |i| {
                if (arg.allow_null) try writer.writeAll("?*") else try writer.writeByte('*');
                try printAbsolute(side, writer, i);
            } else {
                if (arg.allow_null) try writer.writeByte('?');
                try writer.writeAll("*common.Object");
            },
            .array => {
                if (arg.allow_null) try writer.writeByte('?');
                try writer.writeAll("*common.Array");
            },
            .fd => try writer.writeAll("os.fd_t"),
        }
    }
};

const Enum = struct {
    name: []const u8,
    since: u32,
    entries: std.ArrayList(Entry),
    bitfield: bool,

    fn parse(allocator: *mem.Allocator, parser: *xml.Parser) !Enum {
        var name: ?[]const u8 = null;
        var since: ?u32 = null;
        var entries = std.ArrayList(Entry).init(allocator);
        var bitfield: ?bool = null;

        while (parser.next()) |ev| switch (ev) {
            .open_tag => |tag| {
                // TODO: parse description
                if (mem.eql(u8, tag, "entry"))
                    try entries.append(try Entry.parse(allocator, parser));
            },
            .attribute => |attr| {
                if (mem.eql(u8, attr.name, "name")) {
                    if (name != null) return error.DuplicateName;
                    name = try attr.dupeValue(allocator);
                } else if (mem.eql(u8, attr.name, "since")) {
                    if (since != null) return error.DuplicateSince;
                    since = try std.fmt.parseInt(u32, try attr.dupeValue(allocator), 10);
                } else if (mem.eql(u8, attr.name, "bitfield")) {
                    if (bitfield != null) return error.DuplicateBitfield;
                    if (!attr.valueEql("true") and !attr.valueEql("false")) return error.InvalidBoolValue;
                    bitfield = attr.valueEql("true");
                }
            },
            .close_tag => |tag| if (mem.eql(u8, tag, "enum")) {
                return Enum{
                    .name = name orelse return error.MissingName,
                    .since = since orelse 1,
                    .entries = entries,
                    .bitfield = bitfield orelse false,
                };
            },
            else => {},
        };
        return error.UnexpectedEndOfFile;
    }

    fn emit(e: Enum, writer: anytype) !void {
        try writer.writeAll("pub const ");
        try printIdentifier(writer, case(.title, e.name));
        if (e.bitfield) {
            var entries_emitted: u8 = 0;
            try writer.writeAll(" = packed struct {");
            for (e.entries.items) |entry| {
                const value = entry.intValue();
                if (value != 0 and std.math.isPowerOfTwo(value)) {
                    try printIdentifier(writer, entry.name);
                    if (entries_emitted == 0) {
                        // Align the first field to ensure the entire packed
                        // struct matches the alignment of a u32. This allows
                        // using the packed struct as the field of an extern
                        // struct where a u32 is expected.
                        try writer.writeAll(": bool align(@alignOf(u32)) = false,");
                    } else {
                        try writer.writeAll(": bool = false,");
                    }
                    entries_emitted += 1;
                }
            }
            // Pad to 32 bits
            try writer.print("_: u{} = 0,}};\n", .{32 - entries_emitted});
        } else {
            try writer.writeAll(" = extern enum(c_int) {");
            for (e.entries.items) |entry| {
                try printIdentifier(writer, entry.name);
                try writer.print("= {},", .{entry.value});
            }
            // Always generate non-exhaustive enums to ensure forward compatability
            try writer.writeAll("_,};\n");
        }
    }
};

const Entry = struct {
    name: []const u8,
    since: u32,
    value: []const u8,

    fn parse(allocator: *mem.Allocator, parser: *xml.Parser) !Entry {
        var name: ?[]const u8 = null;
        var since: ?u32 = null;
        var value: ?[]const u8 = null;

        while (parser.next()) |ev| switch (ev) {
            .attribute => |attr| {
                if (mem.eql(u8, attr.name, "name")) {
                    if (name != null) return error.DuplicateName;
                    name = try attr.dupeValue(allocator);
                } else if (mem.eql(u8, attr.name, "since")) {
                    if (since != null) return error.DuplicateSince;
                    since = try std.fmt.parseInt(u32, try attr.dupeValue(allocator), 10);
                } else if (mem.eql(u8, attr.name, "value")) {
                    if (value != null) return error.DuplicateName;
                    value = try attr.dupeValue(allocator);
                }
            },
            .close_tag => |tag| if (mem.eql(u8, tag, "entry")) {
                return Entry{
                    .name = name orelse return error.MissingName,
                    .since = since orelse 1,
                    .value = value orelse return error.MissingValue,
                };
            },
            else => {},
        };
        return error.UnexpectedEndOfFile;
    }

    // Return numeric value of enum entry. Can be base 10 and hexadecimal notation.
    fn intValue(e: Entry) u32 {
        return std.fmt.parseInt(u32, e.value, 10) catch blk: {
            const index = mem.indexOfScalar(u8, e.value, 'x').?;
            break :blk std.fmt.parseInt(u32, e.value[index + 1 ..], 16) catch @panic("Can't parse enum entry.");
        };
    }
};

fn prefix(s: []const u8) []const u8 {
    return s[0..mem.indexOfScalar(u8, s, '_').?];
}

fn trimPrefix(s: []const u8) []const u8 {
    return s[mem.indexOfScalar(u8, s, '_').? + 1 ..];
}

var case_buf: [512]u8 = undefined;
fn case(out_case: enum { title, camel }, snake_case: []const u8) []const u8 {
    var i: usize = 0;
    var upper = out_case == .title;
    for (snake_case) |ch| {
        if (ch == '_') {
            upper = true;
            continue;
        }
        case_buf[i] = if (upper) std.ascii.toUpper(ch) else ch;
        i += 1;
        upper = false;
    }
    return case_buf[0..i];
}

fn printAbsolute(side: Side, writer: anytype, interface: []const u8) !void {
    try writer.writeAll(@tagName(side));
    try writer.writeByte('.');
    try printIdentifier(writer, prefix(interface));
    try writer.writeByte('.');
    try printIdentifier(writer, case(.title, trimPrefix(interface)));
}

fn printIdentifier(writer: anytype, identifier: []const u8) !void {
    if (isValidIdentifier(identifier))
        try writer.writeAll(identifier)
    else
        try writer.print("@\"{}\"", .{identifier});
}

fn isValidIdentifier(identifier: []const u8) bool {
    // !keyword [A-Za-z_] [A-Za-z0-9_]*
    if (identifier.len == 0) return false;
    for (identifier) |ch, i| switch (ch) {
        'A'...'Z', 'a'...'z', '_' => {},
        '0'...'9' => if (i == 0) return false,
        else => return false,
    };
    return std.zig.Token.getKeyword(identifier) == null;
}

test "parsing" {
    const testing = std.testing;
    var arena = std.heap.ArenaAllocator.init(testing.allocator);
    defer arena.deinit();

    const protocol = try Protocol.parseXML(&arena.allocator, @embedFile("../protocol/wayland.xml"));

    testing.expectEqualSlices(u8, "wayland", protocol.name);
    testing.expectEqual(@as(usize, 22), protocol.interfaces.items.len);

    {
        const wl_display = protocol.interfaces.items[0];
        testing.expectEqualSlices(u8, "wl_display", wl_display.name);
        testing.expectEqual(@as(u32, 1), wl_display.version);
        testing.expectEqual(@as(usize, 2), wl_display.requests.items.len);
        testing.expectEqual(@as(usize, 2), wl_display.events.items.len);
        testing.expectEqual(@as(usize, 1), wl_display.enums.items.len);

        {
            const sync = wl_display.requests.items[0];
            testing.expectEqualSlices(u8, "sync", sync.name);
            testing.expectEqual(@as(u32, 1), sync.since);
            testing.expectEqual(@as(usize, 1), sync.args.items.len);
            {
                const callback = sync.args.items[0];
                testing.expectEqualSlices(u8, "callback", callback.name);
                testing.expect(callback.kind == .new_id);
                testing.expectEqualSlices(u8, "wl_callback", callback.kind.new_id.?);
                testing.expectEqual(false, callback.allow_null);
                testing.expectEqual(@as(?[]const u8, null), callback.enum_name);
            }
            testing.expect(sync.kind == .constructor);
        }

        {
            const error_event = wl_display.events.items[0];
            testing.expectEqualSlices(u8, "error", error_event.name);
            testing.expectEqual(@as(u32, 1), error_event.since);
            testing.expectEqual(@as(usize, 3), error_event.args.items.len);
            {
                const object_id = error_event.args.items[0];
                testing.expectEqualSlices(u8, "object_id", object_id.name);
                testing.expectEqual(Arg.Type{ .object = null }, object_id.kind);
                testing.expectEqual(false, object_id.allow_null);
                testing.expectEqual(@as(?[]const u8, null), object_id.enum_name);
            }
            {
                const code = error_event.args.items[1];
                testing.expectEqualSlices(u8, "code", code.name);
                testing.expectEqual(Arg.Type.uint, code.kind);
                testing.expectEqual(false, code.allow_null);
                testing.expectEqual(@as(?[]const u8, null), code.enum_name);
            }
            {
                const message = error_event.args.items[2];
                testing.expectEqualSlices(u8, "message", message.name);
                testing.expectEqual(Arg.Type.string, message.kind);
                testing.expectEqual(false, message.allow_null);
                testing.expectEqual(@as(?[]const u8, null), message.enum_name);
            }
        }

        {
            const error_enum = wl_display.enums.items[0];
            testing.expectEqualSlices(u8, "error", error_enum.name);
            testing.expectEqual(@as(u32, 1), error_enum.since);
            testing.expectEqual(@as(usize, 4), error_enum.entries.items.len);
            {
                const invalid_object = error_enum.entries.items[0];
                testing.expectEqualSlices(u8, "invalid_object", invalid_object.name);
                testing.expectEqual(@as(u32, 1), invalid_object.since);
                testing.expectEqualSlices(u8, "0", invalid_object.value);
            }
            {
                const invalid_method = error_enum.entries.items[1];
                testing.expectEqualSlices(u8, "invalid_method", invalid_method.name);
                testing.expectEqual(@as(u32, 1), invalid_method.since);
                testing.expectEqualSlices(u8, "1", invalid_method.value);
            }
            {
                const no_memory = error_enum.entries.items[2];
                testing.expectEqualSlices(u8, "no_memory", no_memory.name);
                testing.expectEqual(@as(u32, 1), no_memory.since);
                testing.expectEqualSlices(u8, "2", no_memory.value);
            }
            {
                const implementation = error_enum.entries.items[3];
                testing.expectEqualSlices(u8, "implementation", implementation.name);
                testing.expectEqual(@as(u32, 1), implementation.since);
                testing.expectEqualSlices(u8, "3", implementation.value);
            }
            testing.expectEqual(false, error_enum.bitfield);
        }
    }

    {
        const wl_data_offer = protocol.interfaces.items[7];
        testing.expectEqualSlices(u8, "wl_data_offer", wl_data_offer.name);
        testing.expectEqual(@as(u32, 3), wl_data_offer.version);
        testing.expectEqual(@as(usize, 5), wl_data_offer.requests.items.len);
        testing.expectEqual(@as(usize, 3), wl_data_offer.events.items.len);
        testing.expectEqual(@as(usize, 1), wl_data_offer.enums.items.len);

        {
            const accept = wl_data_offer.requests.items[0];
            testing.expectEqualSlices(u8, "accept", accept.name);
            testing.expectEqual(@as(u32, 1), accept.since);
            testing.expectEqual(@as(usize, 2), accept.args.items.len);
            {
                const serial = accept.args.items[0];
                testing.expectEqualSlices(u8, "serial", serial.name);
                testing.expectEqual(Arg.Type.uint, serial.kind);
                testing.expectEqual(false, serial.allow_null);
            }
            {
                const mime_type = accept.args.items[1];
                testing.expectEqualSlices(u8, "mime_type", mime_type.name);
                testing.expectEqual(Arg.Type.string, mime_type.kind);
                testing.expectEqual(true, mime_type.allow_null);
            }
        }
    }
}
